@@ -1,4 +1,4 @@
-// Copyright (C) 2013-2015 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2013-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -11,4 +11,6 @@
 /// tests.  See callout_common.cc for details.
 
 static const int LIBRARY_NUMBER = 2;
-#include "callout_library_common.h"
+#include <config.h>
+#include <dhcp4/tests/callout_library_common.h>
+
