@@ -1,17 +1,17 @@
-# Copyright (C) 2014-2016 Internet Systems Consortium, Inc. ("ISC")
+# Copyright (C) 2014-2017 Internet Systems Consortium, Inc. ("ISC")
 #
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
 # Path to the temporary configuration file.
-CFG_FILE=/home/wlodek/dev/kea_110_1/src/bin/dhcp6/tests/test_config.json
+CFG_FILE=/home/wlodek/dev/kea/src/bin/dhcp6/tests/test_config.json
 # Path to the Kea log file.
-LOG_FILE=/home/wlodek/dev/kea_110_1/src/bin/dhcp6/tests/test.log
+LOG_FILE=/home/wlodek/dev/kea/src/bin/dhcp6/tests/test.log
 # Path to the Kea lease file.
-LEASE_FILE=/home/wlodek/dev/kea_110_1/src/bin/dhcp6/tests/test_leases.csv
+LEASE_FILE=/home/wlodek/dev/kea/src/bin/dhcp6/tests/test_leases.csv
 # Expected version
-EXPECTED_VERSION="1.1.0"
+EXPECTED_VERSION="1.5.0"
 # Kea configuration to be stored in the configuration file.
 CONFIG="{
     \"Dhcp6\":
@@ -59,12 +59,53 @@ CONFIG="{
         ]
     }
 }"
+# Invalid configuration (syntax error) to check that Kea can check syntax.
+# This config has following errors:
+# - it should be interfaces-config/interfaces, not interfaces
+# - it should be subnet6/pools, no subnet6/pool
+CONFIG_BAD_SYNTAX="{
+    \"Dhcp6\":
+    {
+        \"interfaces\": [ ],
+        \"preferred-lifetime\": 3000,
+        \"valid-lifetime\": 4000,
+        \"renew-timer\": 1000,
+        \"rebind-timer\": 2000,
+        \"lease-database\":
+        {
+            \"type\": \"memfile\",
+            \"persist\": false
+        },
+        \"subnet6\": [
+        {
+            \"subnet\": \"2001:db8:1::/64\",
+            \"pool\": [ { \"pool\": \"2001:db8:1::10-2001:db8:1::100\" } ]
+        } ]
+    },
+
+    \"Logging\":
+    {
+        \"loggers\": [
+        {
+            \"name\": \"kea-dhcp6\",
+            \"output_options\": [
+                {
+                    \"output\": \"$LOG_FILE\"
+                }
+            ],
+            \"severity\": \"INFO\"
+        }
+        ]
+    }
+}"
 # Invalid configuration (negative preferred-lifetime) to check that Kea
 # gracefully handles reconfiguration errors.
 CONFIG_INVALID="{
     \"Dhcp6\":
     {
-        \"interfaces\": [ ],
+        \"interfaces-config\": {
+          \"interfaces\": [ ]
+        },
         \"preferred-lifetime\": -3,
         \"valid-lifetime\": 4000,
         \"renew-timer\": 1000,
@@ -97,12 +138,75 @@ CONFIG_INVALID="{
     }
 }"
 
+# This config has bad pool values. The pool it out of scope for the subnet
+# it is defined in. Syntactically the config is correct, though.
+CONFIG_BAD_VALUES="{
+    \"Dhcp6\":
+    {   \"interfaces-config\": {
+          \"interfaces\": [ ]
+        },
+        \"server-id\": {
+          \"type\": \"LLT\",
+          \"persist\": false
+        },
+        \"preferred-lifetime\": 3000,
+        \"valid-lifetime\": 4000,
+        \"renew-timer\": 1000,
+        \"rebind-timer\": 2000,
+        \"lease-database\":
+        {
+            \"type\": \"memfile\",
+            \"name\": \"$LEASE_FILE\",
+            \"persist\": false,
+            \"lfc-interval\": 0
+        },
+        \"subnet6\": [
+        {
+            \"subnet\": \"2001:db8::/64\",
+            \"pools\": [ { \"pool\": \"3000::-3000::ffff\" } ]
+        } ],
+        \"dhcp-ddns\": {
+            \"enable-updates\": true,
+            \"qualifying-suffix\": \"\"
+        }
+    }
+}"
+
+
 # Set the location of the executable.
 bin="kea-dhcp6"
-bin_path=/home/wlodek/dev/kea_110_1/src/bin/dhcp6
+bin_path=/home/wlodek/dev/kea/src/bin/dhcp6
 
 # Import common test library.
-. /home/wlodek/dev/kea_110_1/src/lib/testutils/dhcp_test_lib.sh
+. /home/wlodek/dev/kea/src/lib/testutils/dhcp_test_lib.sh
+
+# This test verifies that syntax checking works properly. This function
+# requires 3 parameters:
+# testname
+# config - string with a content of the config (will be written to a file)
+# exp_code - expected exit code returned by kea (0 - success, 1 - failure)
+syntax_check_test() {
+    local TESTNAME="${1}"
+    local CONFIG="${2}"
+    local EXP_CODE="${3}"
+
+    # Log the start of the test and print test name.
+    test_start $TESTNAME
+    # Remove dangling Kea instances and remove log files.
+    cleanup
+    # Create correct configuration file.
+    create_config "${CONFIG}"
+    # Check it
+    printf "Running command %s.\n" "\"${bin_path}/${bin} -t ${CFG_FILE}\""
+    ${bin_path}/${bin} -t ${CFG_FILE}
+    exit_code=$?
+    if [ ${exit_code} -ne $EXP_CODE ]; then
+        printf "ERROR: expected exit code $EXP_CODE, got ${exit_code}\n"
+        clean_exit 1
+    fi
+
+    test_finish 0
+}
 
 # This test verifies that DHCPv6 can be reconfigured with a SIGHUP signal.
 dynamic_reconfiguration_test() {
@@ -278,7 +382,7 @@ returned %d."
 }
 
 # This test verifies that DHCPv6 can be configured to run lease file cleanup
-# periodially.
+# periodically.
 lfc_timer_test() {
     # Log the start of the test and print test name.
     test_start "dhcpv6_srv.lfc_timer_test"
@@ -286,7 +390,7 @@ lfc_timer_test() {
     cleanup
     # Create a configuration with the LFC enabled, by replacing the section
     # with the lfc-interval and persist parameters.
-    LFC_CONFIG=$(printf "${CONFIG}" | sed -e 's/\"lfc-interval\": 0/\"lfc-interval\": 1/g' \
+    LFC_CONFIG=$(printf "${CONFIG}" | sed -e 's/\"lfc-interval\": 0/\"lfc-interval\": 3/g' \
                         | sed -e 's/\"persist\": false,/\"persist\": true,/g')
     # Create new configuration file.
     create_config "${LFC_CONFIG}"
@@ -321,7 +425,7 @@ lfc_timer_test() {
     sleep 1
 
     # Modify the interval.
-    LFC_CONFIG=$(printf "${CONFIG}" | sed -e 's/\"lfc-interval\": 1/\"lfc-interval\": 2/g')
+    LFC_CONFIG=$(printf "${LFC_CONFIG}" | sed -e 's/\"lfc-interval\": 3/\"lfc-interval\": 4/g')
     # Create new configuration file.
     create_config "${LFC_CONFIG}"
 
@@ -383,4 +487,6 @@ shutdown_test "dhcpv6.sigint_test" 2
 version_test "dhcpv6.version"
 logger_vars_test "dhcpv6.variables"
 lfc_timer_test
-
+syntax_check_test "dhcpv6.syntax_check_success" "${CONFIG}" 0
+syntax_check_test "dhcpv6.syntax_check_bad_syntax" "${CONFIG_BAD_SYNTAX}" 1
+syntax_check_test "dhcpv6.syntax_check_bad_values" "${CONFIG_BAD_VALUES}" 1
