@@ -1,4 +1,4 @@
-// Copyright (C) 2011-2016 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2011-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -9,6 +9,7 @@
 #include <dhcp/dhcp6.h>
 #include <dhcp/libdhcp++.h>
 #include <dhcp/option.h>
+#include <dhcp/option_space.h>
 #include <dhcp/option_vendor_class.h>
 #include <dhcp/option_vendor.h>
 #include <dhcp/pkt6.h>
@@ -36,6 +37,19 @@ Pkt6::RelayInfo::RelayInfo()
     peeraddr_(DEFAULT_ADDRESS6), relay_msg_len_(0) {
 }
 
+std::string Pkt6::RelayInfo::toText() const {
+    stringstream tmp;
+    tmp << "msg-type=" << static_cast<int>(msg_type_) << "(" << getName(msg_type_)
+        << "), hop-count=" << static_cast<int>(hop_count_)  << "," << endl
+        << "link-address=" << linkaddr_.toText()
+        << ", peer-address=" << peeraddr_.toText() << ", "
+        << options_.size() << " option(s)" << endl;
+    for (auto option = options_.cbegin(); option != options_.cend(); ++option) {
+        tmp << option->second->toText() << endl;
+    }
+    return (tmp.str());
+}
+
 Pkt6::Pkt6(const uint8_t* buf, uint32_t buf_len, DHCPv6Proto proto /* = UDP */)
    :Pkt(buf, buf_len, DEFAULT_ADDRESS6, DEFAULT_ADDRESS6, 0, 0),
     proto_(proto), msg_type_(0) {
@@ -267,7 +281,7 @@ Pkt6::pack() {
 void
 Pkt6::packUDP() {
     try {
-        // Make sure that the buffer is empty before we start writting to it.
+        // Make sure that the buffer is empty before we start writing to it.
         buffer_out_.clear();
 
         // is this a relayed packet?
@@ -284,7 +298,7 @@ Pkt6::packUDP() {
             for (vector<RelayInfo>::iterator relay = relay_info_.begin();
                  relay != relay_info_.end(); ++relay) {
 
-                // build relay-forw/relay-repl header (see RFC3315, section 7)
+                // build relay-forw/relay-repl header (see RFC 8415, section 9)
                 buffer_out_.writeUint8(relay->msg_type_);
                 buffer_out_.writeUint8(relay->hop_count_);
                 buffer_out_.writeData(&(relay->linkaddr_.toBytes()[0]),
@@ -313,7 +327,7 @@ Pkt6::packUDP() {
 
         }
 
-        // DHCPv6 header: message-type (1 octect) + transaction id (3 octets)
+        // DHCPv6 header: message-type (1 octet) + transaction id (3 octets)
         buffer_out_.writeUint8(msg_type_);
         // store 3-octet transaction-id
         buffer_out_.writeUint8( (transid_ >> 16) & 0xff );
@@ -368,7 +382,7 @@ Pkt6::unpackUDP() {
     case DHCPV6_INFORMATION_REQUEST:
     case DHCPV6_DHCPV4_QUERY:
     case DHCPV6_DHCPV4_RESPONSE:
-    default: // assume that uknown messages are not using relay format
+    default: // assume that unknown messages are not using relay format
         {
             return (unpackMsg(data_.begin(), data_.end()));
         }
@@ -403,7 +417,7 @@ Pkt6::unpackMsg(OptionBuffer::const_iterator begin,
 
     // If custom option parsing function has been set, use this function
     // to parse options. Otherwise, use standard function from libdhcp.
-    size_t offset = LibDHCP::unpackOptions6(opt_buffer, "dhcp6", options_);
+    size_t offset = LibDHCP::unpackOptions6(opt_buffer, DHCP6_OPTION_SPACE, options_);
 
     // If offset is not equal to the size, then something is wrong here. We
     // either parsed past input buffer (bug in our code) or we haven't parsed
@@ -428,7 +442,7 @@ Pkt6::unpackRelayMsg() {
     // we use offset + bufsize, because we want to avoid creating unnecessary
     // copies. There may be up to 32 relays. While using InputBuffer would
     // be probably a bit cleaner, copying data up to 32 times is unacceptable
-    // price here. Hence a single buffer with offets and lengths.
+    // price here. Hence a single buffer with offsets and lengths.
     size_t bufsize = data_.size();
     size_t offset = 0;
 
@@ -453,7 +467,7 @@ Pkt6::unpackRelayMsg() {
 
         // If custom option parsing function has been set, use this function
         // to parse options. Otherwise, use standard function from libdhcp.
-        LibDHCP::unpackOptions6(opt_buffer, "dhcp6", relay.options_,
+        LibDHCP::unpackOptions6(opt_buffer, DHCP6_OPTION_SPACE, relay.options_,
                                 &relay_msg_offset, &relay_msg_len);
 
         /// @todo: check that each option appears at most once
@@ -524,6 +538,9 @@ Pkt6::getMACFromDUID() {
     }
 
     uint8_t hlen = opt_duid->getData().size();
+    if (!hlen) {
+        return (mac);
+    }
     vector<uint8_t> hw_addr(hlen, 0);
     std::vector<unsigned char> duid_data = opt_duid->getData();
 
@@ -585,7 +602,7 @@ Pkt6::makeLabel(const DuidPtr duid, const HWAddrPtr& hwaddr) {
     label << "duid=[" << (duid ? duid->toText() : "no info")
           << "]";
 
-    // HW address is typically not carried in the DHCPv6 mmessages
+    // HW address is typically not carried in the DHCPv6 messages
     // and can be extracted using various, but not fully reliable,
     // techniques. If it is not present, don't print anything.
     if (hwaddr) {
@@ -606,16 +623,31 @@ Pkt6::getLabel() const {
 std::string
 Pkt6::toText() const {
     stringstream tmp;
+
+    // First print the basics
     tmp << "localAddr=[" << local_addr_ << "]:" << local_port_
-        << " remoteAddr=[" << remote_addr_
-        << "]:" << remote_port_ << endl;
-    tmp << "msgtype=" << static_cast<int>(msg_type_) << ", transid=0x" <<
+        << " remoteAddr=[" << remote_addr_ << "]:" << remote_port_ << endl;
+    tmp << "msgtype=" << static_cast<int>(msg_type_) << "(" << getName(msg_type_)
+        << "), transid=0x" <<
         hex << transid_ << dec << endl;
+
+    // Then print the options
     for (isc::dhcp::OptionCollection::const_iterator opt=options_.begin();
          opt != options_.end();
          ++opt) {
         tmp << opt->second->toText() << std::endl;
     }
+
+    // Finally, print the relay information (if present)
+    if (!relay_info_.empty()) {
+        tmp << relay_info_.size() << " relay(s):" << endl;
+        int cnt = 0;
+        for (auto relay = relay_info_.cbegin(); relay != relay_info_.cend(); ++relay) {
+            tmp << "relay[" << cnt++ << "]: " << relay->toText();
+        }
+    } else {
+        tmp << "No relays traversed." << endl;
+    }
     return tmp.str();
 }
 
@@ -765,6 +797,12 @@ void Pkt6::copyRelayInfo(const Pkt6Ptr& question) {
             info.options_.insert(make_pair(opt->getType(), opt));
         }
 
+        // Same for relay-source-port option
+        opt = question->getNonCopiedRelayOption(D6O_RELAY_SOURCE_PORT, i);
+        if (opt) {
+            info.options_.insert(make_pair(opt->getType(), opt));
+        }
+
         /// @todo: Implement support for ERO (Echo Request Option, RFC4994)
 
         // Add this relay-forw info (client's message) to our relay-repl
