@@ -1,4 +1,4 @@
-// Copyright (C) 2013-2016 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2013-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -18,8 +18,10 @@ namespace isc {
 namespace dhcp {
 
 PktFilterInet::PktFilterInet()
-    : control_buf_len_(CMSG_SPACE(sizeof(struct in6_pktinfo))),
-      control_buf_(new char[control_buf_len_])
+    : recv_control_buf_len_(CMSG_SPACE(sizeof(struct in6_pktinfo))),
+      send_control_buf_len_(CMSG_SPACE(sizeof(struct in6_pktinfo))),
+      recv_control_buf_(new char[recv_control_buf_len_]),
+      send_control_buf_(new char[send_control_buf_len_])
 {
 }
 
@@ -112,7 +114,7 @@ PktFilterInet::receive(Iface& iface, const SocketInfo& socket_info) {
     struct sockaddr_in from_addr;
     uint8_t buf[IfaceMgr::RCVBUFSIZE];
 
-    memset(&control_buf_[0], 0, control_buf_len_);
+    memset(&recv_control_buf_[0], 0, recv_control_buf_len_);
     memset(&from_addr, 0, sizeof(from_addr));
 
     // Initialize our message header structure.
@@ -135,8 +137,8 @@ PktFilterInet::receive(Iface& iface, const SocketInfo& socket_info) {
     // previously asked the kernel to give us packet
     // information (when we initialized the interface), so we
     // should get the destination address from that.
-    m.msg_control = &control_buf_[0];
-    m.msg_controllen = control_buf_len_;
+    m.msg_control = &recv_control_buf_[0];
+    m.msg_controllen = recv_control_buf_len_;
 
     int result = recvmsg(socket_info.sockfd_, &m, 0);
     if (result < 0) {
@@ -211,7 +213,7 @@ PktFilterInet::receive(Iface& iface, const SocketInfo& socket_info) {
 int
 PktFilterInet::send(const Iface&, uint16_t sockfd,
                     const Pkt4Ptr& pkt) {
-    memset(&control_buf_[0], 0, control_buf_len_);
+    memset(&send_control_buf_[0], 0, send_control_buf_len_);
 
     // Set the target address we're sending to.
     sockaddr_in to;
@@ -247,16 +249,29 @@ PktFilterInet::send(const Iface&, uint16_t sockfd,
     // We have to create a "control message", and set that to
     // define the IPv4 packet information. We set the source address
     // to handle correctly interfaces with multiple addresses.
-    m.msg_control = &control_buf_[0];
-    m.msg_controllen = control_buf_len_;
+    m.msg_control = &send_control_buf_[0];
+    m.msg_controllen = send_control_buf_len_;
     struct cmsghdr* cmsg = CMSG_FIRSTHDR(&m);
     cmsg->cmsg_level = IPPROTO_IP;
     cmsg->cmsg_type = IP_PKTINFO;
     cmsg->cmsg_len = CMSG_LEN(sizeof(struct in_pktinfo));
     struct in_pktinfo* pktinfo =(struct in_pktinfo *)CMSG_DATA(cmsg);
     memset(pktinfo, 0, sizeof(struct in_pktinfo));
-    pktinfo->ipi_ifindex = pkt->getIndex();
-    pktinfo->ipi_spec_dst.s_addr = htonl(pkt->getLocalAddr().toUint32()); // set the source IP address
+
+    // In some cases the index of the outbound interface is not set. This
+    // is a matter of configuration. When the server is configured to
+    // determine the outbound interface based on routing information,
+    // the index is left unset (negative).
+    if (pkt->indexSet()) {
+        pktinfo->ipi_ifindex = pkt->getIndex();
+    }
+
+    // When the DHCP server is using routing to determine the outbound
+    // interface, the local address is also left unset.
+    if (!pkt->getLocalAddr().isV4Zero()) {
+        pktinfo->ipi_spec_dst.s_addr = htonl(pkt->getLocalAddr().toUint32());
+    }
+
     m.msg_controllen = CMSG_SPACE(sizeof(struct in_pktinfo));
 #endif
 
