@@ -1,4 +1,4 @@
-// Copyright (C) 2014-2015 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2014-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -8,11 +8,14 @@
 #include <dhcp/dhcp4.h>
 #include <dhcp/tests/iface_mgr_test_config.h>
 #include <dhcpsrv/cfg_iface.h>
+#include <testutils/test_to_element.h>
 #include <gtest/gtest.h>
 
 using namespace isc;
 using namespace isc::dhcp;
 using namespace isc::dhcp::test;
+using namespace isc::test;
+using namespace isc::data;
 
 namespace {
 
@@ -181,6 +184,67 @@ TEST_F(CfgIfaceTest, multipleAddressesSameInterfaceV4) {
     EXPECT_TRUE(socketOpen("eth1", "192.0.2.5"));
 }
 
+// This test checks that it is possible to specify the loopback interface.
+TEST_F(CfgIfaceTest, explicitLoopbackV4) {
+    CfgIface cfg;
+    ASSERT_NO_THROW(cfg.use(AF_INET, "lo"));
+
+    // Use UDP sockets
+    ASSERT_NO_THROW(cfg.useSocketType(AF_INET, CfgIface::SOCKET_UDP));
+
+    // Open sockets on specified interfaces and addresses.
+    cfg.openSockets(AF_INET, DHCP4_SERVER_PORT);
+
+    EXPECT_TRUE(socketOpen("lo", "127.0.0.1"));
+
+    // Close all sockets and make sure they are really closed.
+    cfg.closeSockets();
+    ASSERT_FALSE(socketOpen("lo", "127.0.0.1"));
+
+    // Reset configuration.
+    cfg.reset();
+
+    // Retry with wirdcard
+    ASSERT_NO_THROW(cfg.use(AF_INET, "*"));
+    ASSERT_NO_THROW(cfg.use(AF_INET, "lo"));
+    ASSERT_NO_THROW(cfg.useSocketType(AF_INET, CfgIface::SOCKET_UDP));
+    cfg.openSockets(AF_INET, DHCP4_SERVER_PORT);
+    // It is now allowed to use loopback, even with wildcard.
+    EXPECT_TRUE(socketOpen("lo", "127.0.0.1"));
+    cfg.closeSockets();
+    ASSERT_FALSE(socketOpen("lo", "127.0.0.1"));
+
+    // Retry without UDP sockets (lo can be only used with udp sockets)
+    cfg.reset();
+    ASSERT_NO_THROW(cfg.use(AF_INET, "lo"));
+    cfg.openSockets(AF_INET, DHCP4_SERVER_PORT);
+    // No loopback socket
+    EXPECT_FALSE(socketOpen("lo", "127.0.0.1"));
+
+    // Retry with a second interface
+    cfg.reset();
+    ASSERT_NO_THROW(cfg.use(AF_INET, "eth0"));
+    ASSERT_NO_THROW(cfg.use(AF_INET, "lo"));
+    ASSERT_NO_THROW(cfg.useSocketType(AF_INET, CfgIface::SOCKET_UDP));
+    cfg.openSockets(AF_INET, DHCP4_SERVER_PORT);
+    // The logic used to require lo to be the only interface. That constraint
+    // was removed.
+    EXPECT_TRUE(socketOpen("lo", "127.0.0.1"));
+    cfg.closeSockets();
+    EXPECT_FALSE(socketOpen("lo", "127.0.0.1"));
+
+    // Finally with interfaces and addresses
+    cfg.reset();
+    ASSERT_NO_THROW(cfg.use(AF_INET, "eth0/10.0.0.1"));
+    ASSERT_NO_THROW(cfg.use(AF_INET, "lo/127.0.0.1"));
+    ASSERT_NO_THROW(cfg.useSocketType(AF_INET, CfgIface::SOCKET_UDP));
+    cfg.openSockets(AF_INET, DHCP4_SERVER_PORT);
+    // Only loopback is no longer a constraint
+    EXPECT_TRUE(socketOpen("lo", "127.0.0.1"));
+    cfg.closeSockets();
+    EXPECT_FALSE(socketOpen("lo", "127.0.0.1"));
+}
+
 // This test checks that the interface names can be explicitly selected
 // by their names and IPv6 sockets are opened on these interfaces.
 TEST_F(CfgIfaceTest, explicitNamesV6) {
@@ -296,6 +360,57 @@ TEST_F(CfgIfaceTest, invalidValues) {
     ASSERT_THROW(cfg.use(AF_INET6, "*"), DuplicateIfaceName);
 }
 
+// This test checks that it is possible to specify the loopback interface.
+// Note that without a link-local address an unicast address is required.
+TEST_F(CfgIfaceTest, explicitLoopbackV6) {
+    CfgIface cfg;
+    ASSERT_NO_THROW(cfg.use(AF_INET6, "lo/::1"));
+
+    // Open sockets on specified interfaces and addresses.
+    cfg.openSockets(AF_INET6, DHCP6_SERVER_PORT);
+
+    EXPECT_TRUE(socketOpen("lo", AF_INET6));
+
+    // Close all sockets and make sure they are really closed.
+    cfg.closeSockets();
+    ASSERT_FALSE(socketOpen("lo", AF_INET6));
+
+    // Reset configuration.
+    cfg.reset();
+
+    // Retry with wirdcard
+    ASSERT_NO_THROW(cfg.use(AF_INET6, "*"));
+    ASSERT_NO_THROW(cfg.use(AF_INET6, "lo/::1"));
+    cfg.openSockets(AF_INET6, DHCP6_SERVER_PORT);
+    // The logic used to require lo to be used only on its own, not with a
+    // wildcard. That constraint was removed.
+    EXPECT_TRUE(socketOpen("lo", AF_INET6));
+    cfg.closeSockets();
+    ASSERT_FALSE(socketOpen("lo", AF_INET6));
+
+    // Retry with a second interface
+    cfg.reset();
+    ASSERT_NO_THROW(cfg.use(AF_INET6, "eth0"));
+    ASSERT_NO_THROW(cfg.use(AF_INET6, "lo/::1"));
+    cfg.openSockets(AF_INET6, DHCP6_SERVER_PORT);
+    // The logic used to require lo to be used only on its own, not with a
+    // wildcard. That constraint was removed.
+    EXPECT_TRUE(socketOpen("lo", AF_INET6));
+    cfg.closeSockets();
+    ASSERT_FALSE(socketOpen("lo", AF_INET6));
+
+    // Finally with interfaces and addresses
+    cfg.reset();
+    ASSERT_NO_THROW(cfg.use(AF_INET6, "eth0/2001:db8:1::1"));
+    ASSERT_NO_THROW(cfg.use(AF_INET6, "lo/::1"));
+    cfg.openSockets(AF_INET6, DHCP6_SERVER_PORT);
+    // The logic used to require lo to be used only on its own, not with a
+    // wildcard. That constraint was removed.
+    EXPECT_TRUE(socketOpen("lo", AF_INET6));
+    cfg.closeSockets();
+    ASSERT_FALSE(socketOpen("lo", AF_INET6));
+}
+
 // Test that the equality and inequality operators work fine for CfgIface.
 TEST_F(CfgIfaceTest, equality) {
     CfgIface cfg1;
@@ -342,7 +457,7 @@ TEST_F(CfgIfaceTest, equality) {
     EXPECT_FALSE(cfg1 == cfg2);
     EXPECT_TRUE(cfg1 != cfg2);
 
-    // Finally, both are equal as they use wildacard.
+    // Finally, both are equal as they use wildcard.
     cfg2.use(AF_INET, "*");
     EXPECT_TRUE(cfg1 == cfg2);
     EXPECT_FALSE(cfg1 != cfg2);
@@ -358,6 +473,41 @@ TEST_F(CfgIfaceTest, equality) {
     EXPECT_FALSE(cfg1 != cfg2);
 }
 
+// This test verifies that it is possible to unparse the interface config.
+TEST_F(CfgIfaceTest, unparse) {
+    CfgIface cfg4;
+
+    // Add things in it
+    EXPECT_NO_THROW(cfg4.use(AF_INET, "*"));
+    EXPECT_NO_THROW(cfg4.use(AF_INET, "eth0"));
+    EXPECT_NO_THROW(cfg4.use(AF_INET, "eth1/192.0.2.3"));
+    std::string comment = "{ \"comment\": \"foo\", \"bar\": 1 }";
+    EXPECT_NO_THROW(cfg4.setContext(Element::fromJSON(comment)));
+    
+    // Check unparse
+    std::string expected =
+        "{ \"comment\": \"foo\", "
+        "\"interfaces\": [ \"*\", \"eth0\", \"eth1/192.0.2.3\" ], "
+        "\"re-detect\": false, "
+        "\"user-context\": { \"bar\": 1 } }";
+    runToElementTest<CfgIface>(expected, cfg4);
+
+    // Now check IPv6
+    CfgIface cfg6;
+    EXPECT_NO_THROW(cfg6.use(AF_INET6, "*"));
+    EXPECT_NO_THROW(cfg6.use(AF_INET6, "eth1"));
+    EXPECT_NO_THROW(cfg6.use(AF_INET6, "eth0/2001:db8:1::1"));
+    comment = "{ \"comment\": \"bar\", \"foo\": 2 }";
+    EXPECT_NO_THROW(cfg6.setContext(Element::fromJSON(comment)));
+
+    expected =
+        "{ \"comment\": \"bar\", "
+        "\"interfaces\": [ \"*\", \"eth1\", \"eth0/2001:db8:1::1\" ], "
+        "\"re-detect\": false, "
+        "\"user-context\": { \"foo\": 2 } }";
+    runToElementTest<CfgIface>(expected, cfg6);
+}
+
 // This test verifies that it is possible to specify the socket
 // type to be used by the DHCPv4 server.
 // This test is enabled on LINUX and BSD only, because the
@@ -372,6 +522,13 @@ TEST(CfgIfaceNoStubTest, useSocketType) {
     // For datagram sockets, the direct traffic is not supported.
     ASSERT_TRUE(!IfaceMgr::instance().isDirectResponseSupported());
 
+    // Check unparse
+    std::string expected = "{\n"
+        " \"interfaces\": [ ],\n"
+        " \"dhcp-socket-type\": \"udp\",\n"
+        " \"re-detect\": false }";
+    runToElementTest<CfgIface>(expected, cfg);
+
     // Select raw sockets.
     ASSERT_NO_THROW(cfg.useSocketType(AF_INET, "raw"));
     EXPECT_EQ("raw", cfg.socketTypeToText());
