@@ -1,4 +1,4 @@
-// Copyright (C) 2012-2016 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2012-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -8,14 +8,19 @@
 #include <dhcp/option4_client_fqdn.h>
 #include <dhcp/option6_client_fqdn.h>
 #include <dhcpsrv/d2_client_mgr.h>
+#include <testutils/test_to_element.h>
 #include <exceptions/exceptions.h>
+#include <util/strutil.h>
 
+#include <boost/algorithm/string.hpp>
 #include <gtest/gtest.h>
 
 using namespace std;
 using namespace isc::asiolink;
 using namespace isc::dhcp;
 using namespace isc::util;
+using namespace isc::test;
+using namespace isc::data;
 using namespace isc;
 
 namespace {
@@ -72,13 +77,14 @@ TEST(D2ClientConfigTest, constructorsAndAccessors) {
     size_t max_queue_size = 2048;
     dhcp_ddns::NameChangeProtocol ncr_protocol = dhcp_ddns::NCR_UDP;
     dhcp_ddns::NameChangeFormat ncr_format = dhcp_ddns::FMT_JSON;
-    bool always_include_fqdn = true;
     bool override_no_update = true;
     bool override_client_update = true;
     D2ClientConfig::ReplaceClientNameMode replace_client_name_mode = D2ClientConfig::
                                                                      RCM_WHEN_PRESENT;
     std::string generated_prefix = "the_prefix";
     std::string qualifying_suffix = "the.suffix.";
+    std::string hostname_char_set = "[^A-Z]";
+    std::string hostname_char_replacement = "*";
 
     // Verify that we can construct a valid, enabled instance.
     ASSERT_NO_THROW(d2_client_config.reset(new
@@ -90,15 +96,20 @@ TEST(D2ClientConfigTest, constructorsAndAccessors) {
                                                           max_queue_size,
                                                           ncr_protocol,
                                                           ncr_format,
-                                                          always_include_fqdn,
                                                           override_no_update,
                                                           override_client_update,
                                                           replace_client_name_mode,
                                                           generated_prefix,
-                                                          qualifying_suffix)));
-
+                                                          qualifying_suffix,
+                                                          hostname_char_set,
+                                                          hostname_char_replacement)));
     ASSERT_TRUE(d2_client_config);
 
+    // Add user context
+    std::string user_context = "{ \"comment\": \"bar\", \"foo\": 1 }";
+    EXPECT_FALSE(d2_client_config->getContext());
+    d2_client_config->setContext(Element::fromJSON(user_context));
+
     // Verify that the accessors return the expected values.
     EXPECT_EQ(d2_client_config->getEnableUpdates(), enable_updates);
 
@@ -109,7 +120,6 @@ TEST(D2ClientConfigTest, constructorsAndAccessors) {
     EXPECT_EQ(d2_client_config->getMaxQueueSize(), max_queue_size);
     EXPECT_EQ(d2_client_config->getNcrProtocol(), ncr_protocol);
     EXPECT_EQ(d2_client_config->getNcrFormat(), ncr_format);
-    EXPECT_EQ(d2_client_config->getAlwaysIncludeFqdn(), always_include_fqdn);
     EXPECT_EQ(d2_client_config->getOverrideNoUpdate(), override_no_update);
     EXPECT_EQ(d2_client_config->getOverrideClientUpdate(),
               override_client_update);
@@ -117,10 +127,39 @@ TEST(D2ClientConfigTest, constructorsAndAccessors) {
     EXPECT_EQ(d2_client_config->getGeneratedPrefix(), generated_prefix);
     EXPECT_EQ(d2_client_config->getQualifyingSuffix(), qualifying_suffix);
 
+    EXPECT_EQ(d2_client_config->getHostnameCharSet(), hostname_char_set);
+    EXPECT_EQ(d2_client_config->getHostnameCharReplacement(), hostname_char_replacement);
+    EXPECT_TRUE(d2_client_config->getHostnameSanitizer());
+
+    ASSERT_TRUE(d2_client_config->getContext());
+    EXPECT_EQ(d2_client_config->getContext()->str(), user_context);
+
     // Verify that toText called by << operator doesn't bomb.
     ASSERT_NO_THROW(std::cout << "toText test:" << std::endl <<
                     *d2_client_config << std::endl);
 
+    // Verify what toElement returns.
+    std::string expected = "{\n"
+        "\"comment\": \"bar\",\n"
+        "\"enable-updates\": true,\n"
+        "\"server-ip\": \"127.0.0.1\",\n"
+        "\"server-port\": 477,\n"
+        "\"sender-ip\": \"127.0.0.1\",\n"
+        "\"sender-port\": 478,\n"
+        "\"max-queue-size\": 2048,\n"
+        "\"ncr-protocol\": \"UDP\",\n"
+        "\"ncr-format\": \"JSON\",\n"
+        "\"override-no-update\": true,\n"
+        "\"override-client-update\": true,\n"
+        "\"replace-client-name\": \"when-present\",\n"
+        "\"generated-prefix\": \"the_prefix\",\n"
+        "\"qualifying-suffix\": \"the.suffix.\",\n"
+        "\"hostname-char-set\": \"[^A-Z]\",\n"
+        "\"hostname-char-replacement\": \"*\",\n"
+        "\"user-context\": { \"foo\": 1 }\n"
+        "}\n";
+    runToElementTest<D2ClientConfig>(expected, *d2_client_config);
+
     // Verify that constructor does not allow use of NCR_TCP.
     /// @todo obviously this becomes invalid once TCP is supported.
     ASSERT_THROW(d2_client_config.reset(new
@@ -132,12 +171,13 @@ TEST(D2ClientConfigTest, constructorsAndAccessors) {
                                                        max_queue_size,
                                                        dhcp_ddns::NCR_TCP,
                                                        ncr_format,
-                                                       always_include_fqdn,
                                                        override_no_update,
                                                        override_client_update,
                                                        replace_client_name_mode,
                                                        generated_prefix,
-                                                       qualifying_suffix)),
+                                                       qualifying_suffix,
+                                                       hostname_char_set,
+                                                       hostname_char_replacement)),
                  D2ClientError);
 
     /// @todo if additional validation is added to ctor, this test needs to
@@ -156,16 +196,16 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(ref_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(ref_config);
 
     // Check a configuration that is identical to reference configuration.
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_TRUE(*ref_config == *test_config);
     EXPECT_FALSE(*ref_config != *test_config);
@@ -174,8 +214,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(false,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -184,8 +224,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     test_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -194,8 +234,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 333, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -204,8 +244,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, test_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -214,8 +254,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 333, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -224,18 +264,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 2048,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
-    ASSERT_TRUE(test_config);
-    EXPECT_FALSE(*ref_config == *test_config);
-    EXPECT_TRUE(*ref_config != *test_config);
-
-    // Check a configuration that differs only by always_include_fqdn.
-    ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
-                    ref_address, 477, ref_address, 478, 1024,
-                    dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    false, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -244,8 +274,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, false, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    false, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -254,8 +284,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, false, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "suf-fix")));
+                    true, false, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -264,8 +294,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_NEVER,
-                    "pre-fix", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_NEVER,
+                    "pre-fix", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -274,8 +304,8 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "bogus", "suf-fix")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "bogus", "suf-fix", "[^A-Z]", "*")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -284,8 +314,28 @@ TEST(D2ClientConfigTest, equalityOperator) {
     ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
                     ref_address, 477, ref_address, 478, 1024,
                     dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                    true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                    "pre-fix", "bogus")));
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "bogus", "[^A-Z]", "*")));
+    ASSERT_TRUE(test_config);
+    EXPECT_FALSE(*ref_config == *test_config);
+    EXPECT_TRUE(*ref_config != *test_config);
+
+    // Check a configuration that differs only by hostname_char_set
+    ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
+                    ref_address, 477, ref_address, 478, 1024,
+                    dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[abc]", "*")));
+    ASSERT_TRUE(test_config);
+    EXPECT_FALSE(*ref_config == *test_config);
+    EXPECT_TRUE(*ref_config != *test_config);
+
+    // Check a configuration that differs only by hostname_char_replacment
+    ASSERT_NO_THROW(test_config.reset(new D2ClientConfig(true,
+                    ref_address, 477, ref_address, 478, 1024,
+                    dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
+                    true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                    "pre-fix", "suf-fix", "[^A-Z]", "x")));
     ASSERT_TRUE(test_config);
     EXPECT_FALSE(*ref_config == *test_config);
     EXPECT_TRUE(*ref_config != *test_config);
@@ -329,8 +379,8 @@ TEST(D2ClientMgr, validConfig) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  true, true, true, D2ClientConfig::RCM_WHEN_PRESENT,
-                                  "pre-fix", "suf-fix")));
+                                  true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                                  "pre-fix", "suf-fix", "", "")));
 
     // Verify that we can assign a new, non-empty configuration.
     ASSERT_NO_THROW(d2_client_mgr->setD2ClientConfig(new_cfg));
@@ -349,6 +399,42 @@ TEST(D2ClientMgr, validConfig) {
     EXPECT_NE(*original_config, *updated_config);
 }
 
+/// @brief Checks passing the D2ClientMgr a valid D2 client configuration
+/// using IPv6 service.
+TEST(D2ClientMgr, ipv6Config) {
+    D2ClientMgrPtr d2_client_mgr;
+
+    // Construct the manager and fetch its initial configuration.
+    ASSERT_NO_THROW(d2_client_mgr.reset(new D2ClientMgr()));
+    D2ClientConfigPtr original_config = d2_client_mgr->getD2ClientConfig();
+    ASSERT_TRUE(original_config);
+
+    // Create a new, enabled config.
+    D2ClientConfigPtr new_cfg;
+    ASSERT_NO_THROW(new_cfg.reset(new D2ClientConfig(true,
+                                  isc::asiolink::IOAddress("::1"), 477,
+                                  isc::asiolink::IOAddress("::1"), 478,
+                                  1024,
+                                  dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
+                                  true, true, D2ClientConfig::RCM_WHEN_PRESENT,
+                                  "pre-fix", "suf-fix", "", "")));
+
+    // Verify that we can assign a new, non-empty configuration.
+    ASSERT_NO_THROW(d2_client_mgr->setD2ClientConfig(new_cfg));
+
+    // Verify that we can fetch the newly assigned configuration.
+    D2ClientConfigPtr updated_config = d2_client_mgr->getD2ClientConfig();
+    ASSERT_TRUE(updated_config);
+    EXPECT_TRUE(updated_config->getEnableUpdates());
+
+    // Make sure convenience method agrees with the updated configuration.
+    EXPECT_TRUE(d2_client_mgr->ddnsEnabled());
+
+    // Make sure the configuration we fetched is the one we assigned,
+    // and not the original configuration.
+    EXPECT_EQ(*new_cfg, *updated_config);
+    EXPECT_NE(*original_config, *updated_config);
+}
 
 /// @brief Tests that analyzeFqdn detects invalid combination of both the
 /// client S and N flags set to true.
@@ -373,8 +459,8 @@ TEST(D2ClientMgr, analyzeFqdnInvalidCombination) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, false, D2ClientConfig::RCM_NEVER,
-                                  "pre-fix", "suf-fix")));
+                                  false, false, D2ClientConfig::RCM_NEVER,
+                                  "pre-fix", "suf-fix", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_TRUE(mgr.ddnsEnabled());
 
@@ -397,8 +483,8 @@ TEST(D2ClientMgr, analyzeFqdnEnabledNoOverrides) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, false, D2ClientConfig::RCM_NEVER,
-                                  "pre-fix", "suf-fix")));
+                                  false, false, D2ClientConfig::RCM_NEVER,
+                                  "pre-fix", "suf-fix", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_TRUE(mgr.ddnsEnabled());
     ASSERT_FALSE(cfg->getOverrideClientUpdate());
@@ -441,8 +527,8 @@ TEST(D2ClientMgr, analyzeFqdnEnabledOverrideNoUpdate) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, true, false, D2ClientConfig::RCM_NEVER,
-                                  "pre-fix", "suf-fix")));
+                                  true, false, D2ClientConfig::RCM_NEVER,
+                                  "pre-fix", "suf-fix", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_TRUE(mgr.ddnsEnabled());
     ASSERT_TRUE(cfg->getOverrideNoUpdate());
@@ -484,8 +570,8 @@ TEST(D2ClientMgr, analyzeFqdnEnabledOverrideClientUpdate) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, true, D2ClientConfig::RCM_NEVER,
-                                  "pre-fix", "suf-fix")));
+                                  false, true, D2ClientConfig::RCM_NEVER,
+                                  "pre-fix", "suf-fix", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_TRUE(mgr.ddnsEnabled());
     ASSERT_FALSE(cfg->getOverrideNoUpdate());
@@ -528,8 +614,8 @@ TEST(D2ClientMgr, adjustFqdnFlagsV4) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, true, false, D2ClientConfig::RCM_NEVER,
-                                  "pre-fix", "suf-fix")));
+                                  true, false, D2ClientConfig::RCM_NEVER,
+                                  "pre-fix", "suf-fix", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_TRUE(mgr.ddnsEnabled());
     ASSERT_TRUE(cfg->getOverrideNoUpdate());
@@ -629,8 +715,8 @@ TEST(D2ClientMgr, qualifyName) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, true, D2ClientConfig::RCM_NEVER,
-                                  "prefix", "suffix.com")));
+                                  false, true, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "suffix.com", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
 
     // Verify that the qualifying suffix gets appended with trailing dot added.
@@ -645,8 +731,8 @@ TEST(D2ClientMgr, qualifyName) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, true, D2ClientConfig::RCM_NEVER,
-                                  "prefix", "suffix.com")));
+                                  false, true, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "suffix.com", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     partial_name = "somehost";
     qualified_name = mgr.qualifyName(partial_name, false); //false means no dot
@@ -659,8 +745,8 @@ TEST(D2ClientMgr, qualifyName) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, true, D2ClientConfig::RCM_NEVER,
-                                  "prefix", ""))); //empty suffix
+                                  false, true, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "", "", ""))); //empty suffix
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     partial_name = "somehost";
     qualified_name = mgr.qualifyName(partial_name, false); //false means no dot
@@ -672,8 +758,8 @@ TEST(D2ClientMgr, qualifyName) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, true, D2ClientConfig::RCM_NEVER,
-                                  "prefix", "hasdot.com.")));
+                                  false, true, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "hasdot.com.", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
 
     // Verify that the qualifying suffix gets appended without dot added.
@@ -691,8 +777,8 @@ TEST(D2ClientMgr, qualifyName) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, true, D2ClientConfig::RCM_NEVER,
-                                  "prefix", "")));
+                                  false, true, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
 
     // Verify that a name with a trailing dot does not get an extraneous
@@ -729,8 +815,8 @@ TEST(D2ClientMgr, generateFqdn) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, true, D2ClientConfig::RCM_NEVER,
-                                  "prefix", "suffix.com")));
+                                  false, true, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "suffix.com", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
 
     // Verify that it works with an IPv4 address.
@@ -763,13 +849,13 @@ TEST(D2ClientMgr, adjustDomainNameV4) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, false, D2ClientConfig::RCM_NEVER,
-                                  "prefix", "suffix.com")));
+                                  false, false, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "suffix.com", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_EQ(D2ClientConfig::RCM_NEVER, cfg->getReplaceClientNameMode());
 
     // replace-client-name is false, client passes in empty fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option4ClientFqdn(0, Option4ClientFqdn::RCODE_CLIENT(),
                                         "", Option4ClientFqdn::PARTIAL));
     response.reset(new Option4ClientFqdn(*request));
@@ -806,13 +892,13 @@ TEST(D2ClientMgr, adjustDomainNameV4) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, false, D2ClientConfig::RCM_WHEN_PRESENT,
-                                  "prefix", "suffix.com")));
+                                  false, false, D2ClientConfig::RCM_WHEN_PRESENT,
+                                  "prefix", "suffix.com", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_EQ(D2ClientConfig::RCM_WHEN_PRESENT, cfg->getReplaceClientNameMode());
 
     // replace-client-name is true, client passes in empty fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option4ClientFqdn(0, Option4ClientFqdn::RCODE_CLIENT(),
                                         "", Option4ClientFqdn::PARTIAL));
     response.reset(new Option4ClientFqdn(*request));
@@ -822,7 +908,7 @@ TEST(D2ClientMgr, adjustDomainNameV4) {
     EXPECT_EQ(Option4ClientFqdn::PARTIAL, response->getDomainNameType());
 
     // replace-client-name is true, client passes in a partial fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option4ClientFqdn(0, Option4ClientFqdn::RCODE_CLIENT(),
                                         "myhost", Option4ClientFqdn::PARTIAL));
     response.reset(new Option4ClientFqdn(*request));
@@ -833,7 +919,7 @@ TEST(D2ClientMgr, adjustDomainNameV4) {
 
 
     // replace-client-name is true, client passes in a full fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option4ClientFqdn(0, Option4ClientFqdn::RCODE_CLIENT(),
                                         "myhost.example.com.",
                                          Option4ClientFqdn::FULL));
@@ -856,13 +942,13 @@ TEST(D2ClientMgr, adjustDomainNameV6) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, false, D2ClientConfig::RCM_NEVER,
-                                  "prefix", "suffix.com")));
+                                  false, false, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "suffix.com", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_EQ(D2ClientConfig::RCM_NEVER, cfg->getReplaceClientNameMode());
 
     // replace-client-name is false, client passes in empty fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option6ClientFqdn(0, "", Option6ClientFqdn::PARTIAL));
     response.reset(new Option6ClientFqdn(*request));
 
@@ -896,13 +982,13 @@ TEST(D2ClientMgr, adjustDomainNameV6) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, false, false, D2ClientConfig::RCM_WHEN_PRESENT,
-                                  "prefix", "suffix.com")));
+                                  false, false, D2ClientConfig::RCM_WHEN_PRESENT,
+                                  "prefix", "suffix.com", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_EQ(D2ClientConfig::RCM_WHEN_PRESENT, cfg->getReplaceClientNameMode());
 
     // replace-client-name is true, client passes in empty fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option6ClientFqdn(0, "", Option6ClientFqdn::PARTIAL));
     response.reset(new Option6ClientFqdn(*request));
 
@@ -911,7 +997,7 @@ TEST(D2ClientMgr, adjustDomainNameV6) {
     EXPECT_EQ(Option6ClientFqdn::PARTIAL, response->getDomainNameType());
 
     // replace-client-name is true, client passes in a partial fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option6ClientFqdn(0, "myhost",
                                         Option6ClientFqdn::PARTIAL));
     response.reset(new Option6ClientFqdn(*request));
@@ -922,7 +1008,7 @@ TEST(D2ClientMgr, adjustDomainNameV6) {
 
 
     // replace-client-name is true, client passes in a full fqdn
-    // reponse domain should be empty/partial.
+    // response domain should be empty/partial.
     request.reset(new Option6ClientFqdn(0, "myhost.example.com.",
                                         Option6ClientFqdn::FULL));
     response.reset(new Option6ClientFqdn(*request));
@@ -946,8 +1032,8 @@ TEST(D2ClientMgr, adjustFqdnFlagsV6) {
                                   isc::asiolink::IOAddress("127.0.0.1"), 478,
                                   1024,
                                   dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
-                                  false, true, false, D2ClientConfig::RCM_NEVER,
-                                  "pre-fix", "suf-fix")));
+                                  true, false, D2ClientConfig::RCM_NEVER,
+                                  "pre-fix", "suf-fix", "", "")));
     ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
     ASSERT_TRUE(mgr.ddnsEnabled());
     ASSERT_TRUE(cfg->getOverrideNoUpdate());
@@ -1028,4 +1114,167 @@ TEST(D2ClientMgr, updateDirectionsV6) {
     // Response S=1, N=1 isn't possible.
 }
 
+/// @brief Tests v4 FQDN name sanitizing
+TEST(D2ClientMgr, sanitizeFqdnV4) {
+    D2ClientMgr mgr;
+
+    // Create enabled configuration.
+    // replace-client-name is false, client passes in empty fqdn
+    D2ClientConfigPtr cfg;
+    ASSERT_NO_THROW(cfg.reset(new D2ClientConfig(true,
+                                  isc::asiolink::IOAddress("127.0.0.1"), 477,
+                                  isc::asiolink::IOAddress("127.0.0.1"), 478,
+                                  1024,
+                                  dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
+                                  false, false, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "suffix.com", "[^A-Za-z0-9-]", "x")));
+    ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
+    ASSERT_EQ(D2ClientConfig::RCM_NEVER, cfg->getReplaceClientNameMode());
+
+    struct Scenario {
+        std::string description_;
+        std::string client_name_;
+        Option4ClientFqdn::DomainNameType name_type_;
+        std::string expected_name_;
+    };
+
+    std::vector<Scenario> scenarios = {
+        {
+        "full FQDN, name unchanged",
+        "One.123.example.com.",
+        Option4ClientFqdn::FULL,
+        "one.123.example.com."
+        },
+        {
+        "partial FQDN, name unchanged, but qualified",
+        "One.123",
+        Option4ClientFqdn::PARTIAL,
+        "one.123.suffix.com."
+        },
+        {
+        "full FQDN, scrubbed",
+        "O#n^e.123.ex&a*mple.com.",
+        Option4ClientFqdn::FULL,
+        "oxnxe.123.exxaxmple.com."
+        },
+        {
+        "partial FQDN, scrubbed and qualified",
+        "One.1+2|3",
+        Option4ClientFqdn::PARTIAL,
+        "one.1x2x3.suffix.com."
+        },
+        {
+        "full FQDN with characters that get escaped",
+        "O n e.123.exa(m)ple.com.",
+        Option4ClientFqdn::FULL,
+        "oxnxe.123.exaxmxple.com."
+        },
+        {
+        "full FQDN with escape sequences",
+        "O\032n\032e.123.example.com.",
+        Option4ClientFqdn::FULL,
+        "oxnxe.123.example.com."
+        }
+    };
+
+    Option4ClientFqdnPtr request;
+    Option4ClientFqdnPtr response;
+    for (auto scenario = scenarios.begin(); scenario != scenarios.end(); ++scenario) {
+        SCOPED_TRACE((*scenario).description_);
+        {
+            request.reset(new Option4ClientFqdn(0, Option4ClientFqdn::RCODE_CLIENT(),
+                                                (*scenario).client_name_,
+                                                (*scenario).name_type_));
+
+            response.reset(new Option4ClientFqdn(*request));
+            mgr.adjustDomainName<Option4ClientFqdn>(*request, *response);
+            EXPECT_EQ((*scenario).expected_name_, response->getDomainName());
+            EXPECT_EQ(Option4ClientFqdn::FULL, response->getDomainNameType());
+        }
+    }
+}
+
+/// @brief Tests v6 FQDN name sanitizing
+/// @todo This test currently verifies that Option6ClientFqdn::DomainName
+/// downcases strings used to construct it.  For some reason, currently
+/// uknown, Option4ClientFqdn preserves the case, while Option6ClientFqdn
+/// downcases it (see setDomainName() in both classes.  See Trac #5700.
+TEST(D2ClientMgr, sanitizeFqdnV6) {
+    D2ClientMgr mgr;
+
+    // Create enabled configuration.
+    // replace-client-name is false, client passes in empty fqdn
+    D2ClientConfigPtr cfg;
+    ASSERT_NO_THROW(cfg.reset(new D2ClientConfig(true,
+                                  isc::asiolink::IOAddress("127.0.0.1"), 477,
+                                  isc::asiolink::IOAddress("127.0.0.1"), 478,
+                                  1024,
+                                  dhcp_ddns::NCR_UDP, dhcp_ddns::FMT_JSON,
+                                  false, false, D2ClientConfig::RCM_NEVER,
+                                  "prefix", "suffix.com", "[^A-Za-z0-9-]", "x")));
+    ASSERT_NO_THROW(mgr.setD2ClientConfig(cfg));
+    ASSERT_EQ(D2ClientConfig::RCM_NEVER, cfg->getReplaceClientNameMode());
+
+    struct Scenario {
+        std::string description_;
+        std::string client_name_;
+        Option6ClientFqdn::DomainNameType name_type_;
+        std::string expected_name_;
+    };
+
+    std::vector<Scenario> scenarios = {
+        {
+            "full FQDN, name unchanged",
+            "One.123.example.com.",
+            Option6ClientFqdn::FULL,
+            "one.123.example.com."
+        },
+        {
+            "partial FQDN, name unchanged, but qualified",
+            "One.123",
+            Option6ClientFqdn::PARTIAL,
+            "one.123.suffix.com."
+        },
+        {
+            "full FQDN, scrubbed",
+            "O#n^e.123.ex&a*mple.com.",
+            Option6ClientFqdn::FULL,
+            "oxnxe.123.exxaxmple.com."
+        },
+        {
+            "partial FQDN, scrubbed and qualified",
+            "One.1+2|3",
+            Option6ClientFqdn::PARTIAL,
+            "one.1x2x3.suffix.com."
+        },
+        {
+            "full FQDN with characters that get escaped",
+            "O n e.123.exa(m)ple.com.",
+            Option6ClientFqdn::FULL,
+            "oxnxe.123.exaxmxple.com."
+        },
+        {
+            "full FQDN with escape sequences",
+            "O\032n\032e.123.example.com.",
+            Option6ClientFqdn::FULL,
+            "oxnxe.123.example.com."
+        }
+    };
+
+    Option6ClientFqdnPtr request;
+    Option6ClientFqdnPtr response;
+    for (auto scenario = scenarios.begin(); scenario != scenarios.end(); ++scenario) {
+        SCOPED_TRACE((*scenario).description_);
+        {
+            request.reset(new Option6ClientFqdn(0, (*scenario).client_name_,
+                                                (*scenario).name_type_));
+
+            response.reset(new Option6ClientFqdn(*request));
+            mgr.adjustDomainName<Option6ClientFqdn>(*request, *response);
+            EXPECT_EQ((*scenario).expected_name_, response->getDomainName());
+            EXPECT_EQ(Option6ClientFqdn::FULL, response->getDomainNameType());
+        }
+    }
+}
+
 } // end of anonymous namespace
