@@ -1,4 +1,4 @@
-// Copyright (C) 2013-2016 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2013-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -9,6 +9,8 @@
 #include <dhcp/duid.h>
 #include <dhcpsrv/lease.h>
 #include <util/pointer_util.h>
+#include <testutils/test_to_element.h>
+#include <cc/data.h>
 #include <gtest/gtest.h>
 #include <vector>
 #include <sstream>
@@ -16,6 +18,8 @@
 using namespace isc;
 using namespace isc::asiolink;
 using namespace isc::dhcp;
+using namespace isc::data;
+using namespace isc::test;
 
 namespace {
 
@@ -46,6 +50,47 @@ Lease4 createLease4(const std::string& hostname, const bool fqdn_fwd,
     return (lease);
 }
 
+/// @brief Tests that an exception is thrown when one of the lease
+/// parameters is missing or invalid during lease parsing.
+///
+/// If the tested parameter is mandatory the test first removes the
+/// specified parameter from the JSON structure and tries to re-create
+/// the lease, expecting an error.
+/// Next, the test sets invalid value for this parameter and also
+/// expected an error.
+///
+/// @tparam LeaseType @c Lease4 or @c Lease6.
+/// @tparam ValueType type of the invalid value to be used for the
+/// specified parameter.
+/// @param lease_as_json a string contains lease in a JSON format.
+/// @param parameter_name name of the lease parameter to be tested.
+/// @param invalid_value invalid parameter value.
+/// @param mandatory boolean value indicating if the parameter is
+/// mandatory.
+template<typename LeaseType, typename ValueType>
+void testInvalidElement(const std::string& lease_as_json,
+                        const std::string& parameter_name,
+                        ValueType invalid_value,
+                        const bool mandatory = true) {
+    ElementPtr lease;
+
+    // If the parameter is mandatory, check that the exception is
+    // thrown if it is missing.
+    if (mandatory) {
+        ASSERT_NO_THROW(lease = Element::fromJSON(lease_as_json));
+        lease->remove(parameter_name);
+        EXPECT_THROW(LeaseType::fromElement(lease), BadValue)
+            << "test failed for " << parameter_name;
+    }
+
+    // Set invalid value and expect an error.
+    ASSERT_NO_THROW(lease = Element::fromJSON(lease_as_json));
+    lease->set(parameter_name, Element::create(invalid_value));
+    EXPECT_THROW(LeaseType::fromElement(lease), BadValue)
+        << "test failed for " << parameter_name
+        << " and invalid value " << invalid_value;
+}
+
 /// @brief Fixture class used in Lease4 testing.
 class Lease4Test : public ::testing::Test {
 public:
@@ -95,10 +140,11 @@ TEST_F(Lease4Test, constructor) {
         EXPECT_TRUE(lease.fqdn_fwd_);
         EXPECT_TRUE(lease.fqdn_rev_);
         EXPECT_EQ(Lease::STATE_DEFAULT, lease.state_);
+        EXPECT_FALSE(lease.getContext());
     }
 }
 
-// This test verfies that copy constructor copies Lease4 fields correctly.
+// This test verifies that copy constructor copies Lease4 fields correctly.
 TEST_F(Lease4Test, copyConstructor) {
 
     // Get current time for the use in Lease4.
@@ -112,6 +158,9 @@ TEST_F(Lease4Test, copyConstructor) {
     // or the default state will be set for the copied lease.
     lease.state_ = Lease::STATE_DECLINED;
 
+    // Set an user context.
+    lease.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+
     // Use copy constructor to copy the lease.
     Lease4 copied_lease(lease);
 
@@ -121,6 +170,11 @@ TEST_F(Lease4Test, copyConstructor) {
     // Client IDs are equal, but they should be in two distinct pointers.
     EXPECT_FALSE(lease.client_id_ == copied_lease.client_id_);
 
+    // User context are equal and point to the same object.
+    ASSERT_TRUE(copied_lease.getContext());
+    EXPECT_TRUE(lease.getContext() == copied_lease.getContext());
+    EXPECT_TRUE(*lease.getContext() == *copied_lease.getContext());
+
     // Hardware addresses are equal, but they should point to two objects,
     // each holding the same data. The content should be equal...
     EXPECT_TRUE(*lease.hwaddr_ == *copied_lease.hwaddr_);
@@ -134,7 +188,7 @@ TEST_F(Lease4Test, copyConstructor) {
     EXPECT_TRUE(lease == copied_lease2);
 }
 
-// This test verfies that the assignment operator copies all Lease4 fields
+// This test verifies that the assignment operator copies all Lease4 fields
 // correctly.
 TEST_F(Lease4Test, operatorAssign) {
 
@@ -149,6 +203,9 @@ TEST_F(Lease4Test, operatorAssign) {
     // or the default state will be set for the copied lease.
     lease.state_ = Lease::STATE_DECLINED;
 
+    // Set an user context.
+    lease.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+
     // Create a default lease.
     Lease4 copied_lease;
     // Use assignment operator to assign new lease.
@@ -160,6 +217,11 @@ TEST_F(Lease4Test, operatorAssign) {
     // Client IDs are equal, but they should be in two distinct pointers.
     EXPECT_FALSE(lease.client_id_ == copied_lease.client_id_);
 
+    // User context are equal and point to the same object.
+    ASSERT_TRUE(copied_lease.getContext());
+    EXPECT_TRUE(lease.getContext() == copied_lease.getContext());
+    EXPECT_TRUE(*lease.getContext() == *copied_lease.getContext());
+
     // Hardware addresses are equal, but they should point to two objects,
     // each holding the same data. The content should be equal...
     EXPECT_TRUE(*lease.hwaddr_ == *copied_lease.hwaddr_);
@@ -249,6 +311,7 @@ TEST_F(Lease4Test, operatorEquals) {
     // Check when the leases are equal.
     Lease4 lease1(ADDRESS, hwaddr_, clientid_, VALID_LIFETIME, current_time, 0,
                   0, SUBNET_ID);
+    lease1.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
 
     // We need to make an explicit copy. Otherwise the second lease will just
     // store a pointer and we'll have two leases pointing to a single HWAddr
@@ -258,6 +321,7 @@ TEST_F(Lease4Test, operatorEquals) {
 
     Lease4 lease2(ADDRESS, hwcopy, clientid_copy, VALID_LIFETIME, current_time,
                   0, 0, SUBNET_ID);
+    lease2.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
     EXPECT_TRUE(lease1 == lease2);
     EXPECT_FALSE(lease1 != lease2);
 
@@ -349,6 +413,20 @@ TEST_F(Lease4Test, operatorEquals) {
     lease2.state_ += 1;
     EXPECT_TRUE(lease1 == lease2);  // Check that the reversion has made the
     EXPECT_FALSE(lease1 != lease2); // ... leases equal
+
+    lease1.setContext(Element::fromJSON("{ \"foobar\": 5678 }"));
+    EXPECT_FALSE(lease1 == lease2);
+    EXPECT_TRUE(lease1 != lease2);
+    lease1.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+    EXPECT_TRUE(lease1 == lease2);  // Check that the reversion has made the
+    EXPECT_FALSE(lease1 != lease2); // ... leases equal
+
+    lease1.setContext(ConstElementPtr());
+    EXPECT_FALSE(lease1 == lease2);
+    EXPECT_TRUE(lease1 != lease2);
+    lease2.setContext(ConstElementPtr());
+    EXPECT_TRUE(lease1 == lease2);  // Check that no user context has mase the
+    EXPECT_FALSE(lease1 != lease2); // ... leases equal
 }
 
 // Verify that the client id can be returned as a vector object and if client
@@ -375,6 +453,9 @@ TEST_F(Lease4Test, hasIdenticalFqdn) {
     Lease4 lease = createLease4("myhost.example.com.", true, true);
     EXPECT_TRUE(lease.hasIdenticalFqdn(createLease4("myhost.example.com.",
                                                      true, true)));
+    // Case insensitive comparison.
+    EXPECT_TRUE(lease.hasIdenticalFqdn(createLease4("myHOst.ExamplE.coM.",
+                                                     true, true)));
     EXPECT_FALSE(lease.hasIdenticalFqdn(createLease4("other.example.com.",
                                                      true, true)));
     EXPECT_FALSE(lease.hasIdenticalFqdn(createLease4("myhost.example.com.",
@@ -393,6 +474,7 @@ TEST_F(Lease4Test, toText) {
     const time_t current_time = 12345678;
     Lease4 lease(IOAddress("192.0.2.3"), hwaddr_, clientid_, 3600, 123,
                  456, current_time, 789);
+    lease.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
 
     std::stringstream expected;
     expected << "Address:       192.0.2.3\n"
@@ -403,13 +485,16 @@ TEST_F(Lease4Test, toText) {
              << "Hardware addr: " << hwaddr_->toText(false) << "\n"
              << "Client id:     " << clientid_->toText() << "\n"
              << "Subnet ID:     789\n"
-             << "State:         default\n";
+             << "State:         default\n"
+             << "User context:  { \"foobar\": 1234 }\n";
 
     EXPECT_EQ(expected.str(), lease.toText());
 
-    // Now let's try with a lease without hardware address and client identifier.
+    // Now let's try with a lease without hardware address, client identifier
+    // and user context.
     lease.hwaddr_.reset();
     lease.client_id_.reset();
+    lease.setContext(ConstElementPtr());
     expected.str("");
     expected << "Address:       192.0.2.3\n"
              << "Valid life:    3600\n"
@@ -423,6 +508,146 @@ TEST_F(Lease4Test, toText) {
     EXPECT_EQ(expected.str(), lease.toText());
 }
 
+// Verify that Lease4 structure can be converted to JSON properly.
+TEST_F(Lease4Test, toElement) {
+
+    const time_t current_time = 12345678;
+    Lease4 lease(IOAddress("192.0.2.3"), hwaddr_, clientid_, 3600, 123,
+                 456, current_time, 789, true, true, "urania.example.org");
+    lease.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+
+    std::string expected = "{"
+        "\"client-id\": \"17:34:e2:ff:09:92:54\","
+        "\"cltt\": 12345678,"
+        "\"fqdn-fwd\": true,"
+        "\"fqdn-rev\": true,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"ip-address\": \"192.0.2.3\","
+        "\"state\": 0,"
+        "\"subnet-id\": 789,"
+        "\"user-context\": { \"foobar\": 1234 },"
+        "\"valid-lft\": 3600 "
+        "}";
+
+    runToElementTest<Lease4>(expected, lease);
+
+    // Now let's try with a lease without client-id and user context.
+    lease.client_id_.reset();
+    lease.setContext(ConstElementPtr());
+
+    expected = "{"
+        "\"cltt\": 12345678,"
+        "\"fqdn-fwd\": true,"
+        "\"fqdn-rev\": true,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"ip-address\": \"192.0.2.3\","
+        "\"state\": 0,"
+        "\"subnet-id\": 789,"
+        "\"valid-lft\": 3600 "
+        "}";
+
+    runToElementTest<Lease4>(expected, lease);
+
+    // And to finish try with a comment.
+    lease.setContext(Element::fromJSON("{ \"comment\": \"a comment\" }"));
+
+    expected = "{"
+        "\"cltt\": 12345678,"
+        "\"comment\": \"a comment\","
+        "\"fqdn-fwd\": true,"
+        "\"fqdn-rev\": true,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"ip-address\": \"192.0.2.3\","
+        "\"state\": 0,"
+        "\"subnet-id\": 789,"
+        "\"valid-lft\": 3600 "
+        "}";
+
+    runToElementTest<Lease4>(expected, lease);
+}
+
+// Verify that the Lease4 can be created from JSON.
+TEST_F(Lease4Test, fromElement) {
+    std::string json = "{"
+        "\"client-id\": \"17:34:e2:ff:09:92:54\","
+        "\"cltt\": 12345678,"
+        "\"fqdn-fwd\": true,"
+        "\"fqdn-rev\": true,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"ip-address\": \"192.0.2.3\","
+        "\"state\": 0,"
+        "\"subnet-id\": 789,"
+        "\"user-context\": { \"foo\": \"bar\" },"
+        "\"valid-lft\": 3600 "
+        "}";
+
+    Lease4Ptr lease;
+    ASSERT_NO_THROW(lease = Lease4::fromElement(Element::fromJSON(json)));
+
+    ASSERT_TRUE(lease);
+
+    EXPECT_EQ("192.0.2.3", lease->addr_.toText());
+    EXPECT_EQ(789, static_cast<uint32_t>(lease->subnet_id_));
+    ASSERT_TRUE(lease->hwaddr_);
+    EXPECT_EQ("hwtype=1 08:00:2b:02:3f:4e", lease->hwaddr_->toText());
+    ASSERT_TRUE(lease->client_id_);
+    EXPECT_EQ("17:34:e2:ff:09:92:54", lease->client_id_->toText());
+    EXPECT_EQ(12345678, lease->cltt_);
+    EXPECT_EQ(3600, lease->valid_lft_);
+    EXPECT_TRUE(lease->fqdn_fwd_);
+    EXPECT_TRUE(lease->fqdn_rev_);
+    EXPECT_EQ("urania.example.org", lease->hostname_);
+    EXPECT_EQ(Lease::STATE_DEFAULT, lease->state_);
+    ASSERT_TRUE(lease->getContext());
+    EXPECT_EQ("{ \"foo\": \"bar\" }", lease->getContext()->str());
+}
+
+// Test that specifying invalid values for a lease or not specifying
+// mandatory lease parameters causes an error while parsing the lease.
+TEST_F(Lease4Test, fromElementInvalidValues) {
+    // Create valid configuration. We use it as a base from which we will
+    // be removing some of the parameters and some values will be selectively
+    // modified.
+    std::string json = "{"
+        "\"client-id\": \"17:34:e2:ff:09:92:54\","
+        "\"cltt\": 12345678,"
+        "\"fqdn-fwd\": true,"
+        "\"fqdn-rev\": true,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"ip-address\": \"192.0.2.3\","
+        "\"state\": 0,"
+        "\"subnet-id\": 789,"
+        "\"valid-lft\": 3600 "
+        "}";
+
+    // Test invalid parameter values and missing parameters.
+    testInvalidElement<Lease4>(json, "client-id", std::string("rock"), false);
+    testInvalidElement<Lease4>(json, "cltt", std::string("xyz"));
+    testInvalidElement<Lease4>(json, "cltt", -1, false);
+    testInvalidElement<Lease4>(json, "fqdn-fwd", 123);
+    testInvalidElement<Lease4>(json, "fqdn-rev", std::string("foo"));
+    testInvalidElement<Lease4, bool>(json, "hostname", true);
+    testInvalidElement<Lease4>(json, "hw-address", "01::00::");
+    testInvalidElement<Lease4>(json, "hw-address", 1234, false);
+    testInvalidElement<Lease4, long int>(json, "ip-address", 0xFF000201);
+    testInvalidElement<Lease4>(json, "ip-address", "2001:db8:1::1", false);
+    testInvalidElement<Lease4>(json, "state", std::string("xyz"));
+    testInvalidElement<Lease4>(json, "state", 1234, false);
+    testInvalidElement<Lease4>(json, "subnet-id", std::string("xyz"));
+    testInvalidElement<Lease4>(json, "subnet-id", -5, false);
+    testInvalidElement<Lease4>(json, "valid-lft", std::string("xyz"));
+    testInvalidElement<Lease4>(json, "valid-lft", -3, false);
+    testInvalidElement<Lease4>(json, "user-context", "[ ]", false);
+    testInvalidElement<Lease4>(json, "user-context", 1234, false);
+    testInvalidElement<Lease4>(json, "user-context", false, false);
+    testInvalidElement<Lease4>(json, "user-context", "foo", false);
+}
+
 // Verify that decline() method properly clears up specific fields.
 TEST_F(Lease4Test, decline) {
 
@@ -450,6 +675,7 @@ TEST_F(Lease4Test, decline) {
     EXPECT_FALSE(lease.fqdn_rev_);
     EXPECT_EQ(Lease::STATE_DECLINED, lease.state_);
     EXPECT_EQ(123, lease.valid_lft_);
+    EXPECT_FALSE(lease.getContext());
 }
 
 // Verify that the lease states are correctly returned in the textual format.
@@ -511,7 +737,7 @@ TEST(Lease6Test, Lease6ConstructorDefault) {
         EXPECT_FALSE(lease->fqdn_fwd_);
         EXPECT_FALSE(lease->fqdn_rev_);
         EXPECT_TRUE(lease->hostname_.empty());
-
+        EXPECT_FALSE(lease->getContext());
     }
 
     // Lease6 must be instantiated with a DUID, not with NULL pointer
@@ -588,6 +814,9 @@ TEST(Lease6Test, operatorEquals) {
     Lease6 lease2(Lease::TYPE_NA, addr, duid, iaid, 100, 200, 50, 80,
                                subnet_id);
 
+    lease1.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+    lease2.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+
     // cltt_ constructs with time(NULL), make sure they are always equal
     lease1.cltt_ = lease2.cltt_;
 
@@ -702,6 +931,20 @@ TEST(Lease6Test, operatorEquals) {
     lease2.state_ += 1;
     EXPECT_TRUE(lease1 == lease2);  // Check that the reversion has made the
     EXPECT_FALSE(lease1 != lease2); // ... leases equal
+
+    lease1.setContext(Element::fromJSON("{ \"foobar\": 5678 }"));
+    EXPECT_FALSE(lease1 == lease2);
+    EXPECT_TRUE(lease1 != lease2);
+    lease1.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+    EXPECT_TRUE(lease1 == lease2);  // Check that the reversion has made the
+    EXPECT_FALSE(lease1 != lease2); // ... leases equal
+
+    lease1.setContext(ConstElementPtr());
+    EXPECT_FALSE(lease1 == lease2);
+    EXPECT_TRUE(lease1 != lease2);
+    lease2.setContext(ConstElementPtr());
+    EXPECT_TRUE(lease1 == lease2);  // Check that no user context has mase the
+    EXPECT_FALSE(lease1 != lease2); // ... leases equal
 }
 
 // Checks if lease expiration is calculated properly
@@ -784,6 +1027,7 @@ TEST(Lease6Test, decline) {
     EXPECT_FALSE(lease.fqdn_rev_);
     EXPECT_EQ(Lease::STATE_DECLINED, lease.state_);
     EXPECT_EQ(123, lease.valid_lft_);
+    EXPECT_FALSE(lease.getContext());
 }
 
 // Verify the behavior of the function which checks FQDN data for equality.
@@ -791,6 +1035,9 @@ TEST(Lease6Test, hasIdenticalFqdn) {
     Lease6 lease = createLease6("myhost.example.com.", true, true);
     EXPECT_TRUE(lease.hasIdenticalFqdn(createLease6("myhost.example.com.",
                                                     true, true)));
+    // Case insensitive comparison.
+    EXPECT_TRUE(lease.hasIdenticalFqdn(createLease6("myHOst.ExamplE.coM.",
+                                                    true, true)));
     EXPECT_FALSE(lease.hasIdenticalFqdn(createLease6("other.example.com.",
                                                      true, true)));
     EXPECT_FALSE(lease.hasIdenticalFqdn(createLease6("myhost.example.com.",
@@ -803,7 +1050,7 @@ TEST(Lease6Test, hasIdenticalFqdn) {
                                                      false, false)));
 }
 
-// Verify that toText() method reports Lease4 structure properly.
+// Verify that toText() method reports Lease6 structure properly.
 TEST(Lease6Test, toText) {
 
     HWAddrPtr hwaddr(new HWAddr(HWADDR, sizeof(HWADDR), HTYPE_ETHER));
@@ -815,6 +1062,7 @@ TEST(Lease6Test, toText) {
                  400, 800, 100, 200, 5678, hwaddr, 128);
     lease.cltt_ = 12345678;
     lease.state_ = Lease::STATE_DECLINED;
+    lease.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
 
     std::stringstream expected;
     expected << "Type:          IA_NA(" << static_cast<int>(Lease::TYPE_NA) << ")\n"
@@ -827,12 +1075,14 @@ TEST(Lease6Test, toText) {
              << "DUID:          00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f\n"
              << "Hardware addr: " << hwaddr->toText(false) << "\n"
              << "Subnet ID:     5678\n"
-             << "State:         declined\n";
+             << "State:         declined\n"
+             << "User context:  { \"foobar\": 1234 }\n";
 
     EXPECT_EQ(expected.str(), lease.toText());
 
-    // Now let's try with a lease without hardware address.
+    // Now let's try with a lease without hardware address and user context.
     lease.hwaddr_.reset();
+    lease.setContext(ConstElementPtr());
     expected.str("");
     expected << "Type:          IA_NA(" << static_cast<int>(Lease::TYPE_NA) << ")\n"
              << "Address:       2001:db8::1\n"
@@ -848,6 +1098,312 @@ TEST(Lease6Test, toText) {
     EXPECT_EQ(expected.str(), lease.toText());
 }
 
+// Verify that Lease6 structure can be converted to JSON properly.
+// This tests an address lease conversion.
+TEST(Lease6Test, toElementAddress) {
+
+    HWAddrPtr hwaddr(new HWAddr(HWADDR, sizeof(HWADDR), HTYPE_ETHER));
+
+    uint8_t llt[] = {0, 1, 2, 3, 4, 5, 6, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf};
+    DuidPtr duid(new DUID(llt, sizeof(llt)));
+
+    Lease6 lease(Lease::TYPE_NA, IOAddress("2001:db8::1"), duid, 123456,
+                 400, 800, 100, 200, 5678, hwaddr, 128);
+    lease.cltt_ = 12345678;
+    lease.state_ = Lease::STATE_DECLINED;
+    lease.hostname_ = "urania.example.org";
+    lease.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+
+    std::string expected = "{"
+        "\"cltt\": 12345678,"
+        "\"duid\": \"00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f\","
+        "\"fqdn-fwd\": false,"
+        "\"fqdn-rev\": false,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"iaid\": 123456,"
+        "\"ip-address\": \"2001:db8::1\","
+        "\"preferred-lft\": 400,"
+        "\"state\": 1,"
+        "\"subnet-id\": 5678,"
+        "\"type\": \"IA_NA\","
+        "\"user-context\": { \"foobar\": 1234 },"
+        "\"valid-lft\": 800"
+        "}";
+    
+    runToElementTest<Lease6>(expected, lease);
+
+    // Now let's try with a lease without hardware address and user context.
+    lease.hwaddr_.reset();
+    lease.setContext(ConstElementPtr());
+
+    expected = "{"
+        "\"cltt\": 12345678,"
+        "\"duid\": \"00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f\","
+        "\"fqdn-fwd\": false,"
+        "\"fqdn-rev\": false,"
+        "\"hostname\": \"urania.example.org\","
+        "\"iaid\": 123456,"
+        "\"ip-address\": \"2001:db8::1\","
+        "\"preferred-lft\": 400,"
+        "\"state\": 1,"
+        "\"subnet-id\": 5678,"
+        "\"type\": \"IA_NA\","
+        "\"valid-lft\": 800"
+        "}";
+    
+    runToElementTest<Lease6>(expected, lease);
+
+    // And to finish try with a comment.
+    lease.setContext(Element::fromJSON("{ \"comment\": \"a comment\" }"));
+
+    expected = "{"
+        "\"cltt\": 12345678,"
+        "\"comment\": \"a comment\","
+        "\"duid\": \"00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f\","
+        "\"fqdn-fwd\": false,"
+        "\"fqdn-rev\": false,"
+        "\"hostname\": \"urania.example.org\","
+        "\"iaid\": 123456,"
+        "\"ip-address\": \"2001:db8::1\","
+        "\"preferred-lft\": 400,"
+        "\"state\": 1,"
+        "\"subnet-id\": 5678,"
+        "\"type\": \"IA_NA\","
+        "\"valid-lft\": 800"
+        "}";
+    
+    runToElementTest<Lease6>(expected, lease);
+}
+
+// Verify that Lease6 structure can be converted to JSON properly.
+// This tests an address lease conversion.
+TEST(Lease6Test, toElementPrefix) {
+
+    HWAddrPtr hwaddr(new HWAddr(HWADDR, sizeof(HWADDR), HTYPE_ETHER));
+
+    uint8_t llt[] = {0, 1, 2, 3, 4, 5, 6, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf};
+    DuidPtr duid(new DUID(llt, sizeof(llt)));
+
+    Lease6 lease(Lease::TYPE_PD, IOAddress("2001:db8::"), duid, 123456,
+                 400, 800, 100, 200, 5678, hwaddr, 56);
+    lease.cltt_ = 12345678;
+    lease.state_ = Lease::STATE_DEFAULT;
+    lease.hostname_ = "urania.example.org";
+    lease.setContext(Element::fromJSON("{ \"foobar\": 1234 }"));
+
+    ElementPtr l = lease.toElement();
+
+    ASSERT_TRUE(l);
+
+    ASSERT_TRUE(l->contains("ip-address"));
+    EXPECT_EQ("2001:db8::", l->get("ip-address")->stringValue());
+
+    ASSERT_TRUE(l->contains("type"));
+    EXPECT_EQ("IA_PD", l->get("type")->stringValue());
+
+    // This is a prefix lease, it must have a prefix length.
+    ASSERT_TRUE(l->contains("prefix-len"));
+    EXPECT_EQ(56, l->get("prefix-len")->intValue());
+
+    ASSERT_TRUE(l->contains("iaid"));
+    EXPECT_EQ(123456, l->get("iaid")->intValue());
+
+    ASSERT_TRUE(l->contains("preferred-lft"));
+    EXPECT_EQ(400, l->get("preferred-lft")->intValue());
+
+    ASSERT_TRUE(l->contains("valid-lft"));
+    EXPECT_EQ(800, l->get("valid-lft")->intValue());
+
+    ASSERT_TRUE(l->contains("duid"));
+    EXPECT_EQ("00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f",
+              l->get("duid")->stringValue());
+
+    ASSERT_TRUE(l->contains("hw-address"));
+    EXPECT_EQ(hwaddr->toText(false), l->get("hw-address")->stringValue());
+
+    ASSERT_TRUE(l->contains("subnet-id"));
+    EXPECT_EQ(5678, l->get("subnet-id")->intValue());
+
+    ASSERT_TRUE(l->contains("state"));
+    EXPECT_EQ(static_cast<int>(Lease::STATE_DEFAULT),
+              l->get("state")->intValue());
+
+    ASSERT_TRUE(l->contains("fqdn-fwd"));
+    EXPECT_FALSE(l->get("fqdn-fwd")->boolValue());
+
+    ASSERT_TRUE(l->contains("fqdn-rev"));
+    EXPECT_FALSE(l->get("fqdn-rev")->boolValue());
+
+    ASSERT_TRUE(l->contains("hostname"));
+    EXPECT_EQ("urania.example.org", l->get("hostname")->stringValue());
+
+    ASSERT_TRUE(l->contains("user-context"));
+    EXPECT_EQ("{ \"foobar\": 1234 }", l->get("user-context")->str());
+
+    // Now let's try with a lease without hardware address or user context.
+    lease.hwaddr_.reset();
+    lease.setContext(ConstElementPtr());
+    l = lease.toElement();
+    EXPECT_FALSE(l->contains("hw-address"));
+    EXPECT_FALSE(l->contains("user-context"));
+    EXPECT_FALSE(l->contains("comment"));
+
+    // And to finish try with a comment.
+    lease.setContext(Element::fromJSON("{ \"comment\": \"a comment\" }"));
+    l = lease.toElement();
+    EXPECT_FALSE(l->contains("hw-address"));
+    EXPECT_FALSE(l->contains("user-context"));
+    ASSERT_TRUE(l->contains("comment"));
+    EXPECT_EQ("a comment", l->get("comment")->stringValue());
+}
+
+// Verify that the IA_NA can be created from JSON.
+TEST(Lease6Test, fromElementNA) {
+    std::string json = "{"
+        "\"cltt\": 12345678,"
+        "\"duid\": \"00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f\","
+        "\"fqdn-fwd\": false,"
+        "\"fqdn-rev\": false,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"iaid\": 123456,"
+        "\"ip-address\": \"2001:db8::1\","
+        "\"preferred-lft\": 400,"
+        "\"state\": 1,"
+        "\"subnet-id\": 5678,"
+        "\"type\": \"IA_NA\","
+        "\"user-context\": { \"foobar\": 1234 },"
+        "\"valid-lft\": 800"
+        "}";
+
+    Lease6Ptr lease;
+    ASSERT_NO_THROW(lease = Lease6::fromElement(Element::fromJSON(json)));
+
+    ASSERT_TRUE(lease);
+
+    EXPECT_EQ("2001:db8::1", lease->addr_.toText());
+    EXPECT_EQ(5678, static_cast<uint32_t>(lease->subnet_id_));
+    ASSERT_TRUE(lease->hwaddr_);
+    EXPECT_EQ("hwtype=1 08:00:2b:02:3f:4e", lease->hwaddr_->toText());
+    EXPECT_EQ(12345678, lease->cltt_);
+    EXPECT_EQ(800, lease->valid_lft_);
+    EXPECT_FALSE(lease->fqdn_fwd_);
+    EXPECT_FALSE(lease->fqdn_rev_);
+    EXPECT_EQ("urania.example.org", lease->hostname_);
+    EXPECT_EQ(Lease::STATE_DECLINED, lease->state_);
+    ASSERT_TRUE(lease->getContext());
+    EXPECT_EQ("{ \"foobar\": 1234 }", lease->getContext()->str());
+
+    // IPv6 specific properties.
+    EXPECT_EQ(Lease::TYPE_NA, lease->type_);
+    EXPECT_EQ(0, lease->prefixlen_);
+    EXPECT_EQ(123456, lease->iaid_);
+    ASSERT_TRUE(lease->duid_);
+    EXPECT_EQ("00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f", lease->duid_->toText());
+    EXPECT_EQ(400, lease->preferred_lft_);
+}
+
+// Verify that the IA_PD can be created from JSON.
+TEST(Lease6Test, fromElementPD) {
+    std::string json = "{"
+        "\"cltt\": 12345678,"
+        "\"duid\": \"00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f\","
+        "\"fqdn-fwd\": false,"
+        "\"fqdn-rev\": false,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"iaid\": 123456,"
+        "\"ip-address\": \"3000::\","
+        "\"preferred-lft\": 400,"
+        "\"prefix-len\": 32,"
+        "\"state\": 0,"
+        "\"subnet-id\": 1234,"
+        "\"type\": \"IA_PD\","
+        "\"valid-lft\": 600"
+        "}";
+
+    Lease6Ptr lease;
+    ASSERT_NO_THROW(lease = Lease6::fromElement(Element::fromJSON(json)));
+
+    ASSERT_TRUE(lease);
+
+    EXPECT_EQ("3000::", lease->addr_.toText());
+    EXPECT_EQ(1234, static_cast<uint32_t>(lease->subnet_id_));
+    ASSERT_TRUE(lease->hwaddr_);
+    EXPECT_EQ("hwtype=1 08:00:2b:02:3f:4e", lease->hwaddr_->toText());
+    EXPECT_EQ(12345678, lease->cltt_);
+    EXPECT_EQ(600, lease->valid_lft_);
+    EXPECT_FALSE(lease->fqdn_fwd_);
+    EXPECT_FALSE(lease->fqdn_rev_);
+    EXPECT_EQ("urania.example.org", lease->hostname_);
+    EXPECT_EQ(Lease::STATE_DEFAULT , lease->state_);
+    EXPECT_FALSE(lease->getContext());
+
+    // IPv6 specific properties.
+    EXPECT_EQ(Lease::TYPE_PD, lease->type_);
+    EXPECT_EQ(32, static_cast<int>(lease->prefixlen_));
+    EXPECT_EQ(123456, lease->iaid_);
+    ASSERT_TRUE(lease->duid_);
+    EXPECT_EQ("00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f", lease->duid_->toText());
+    EXPECT_EQ(400, lease->preferred_lft_);
+}
+
+// Test that specifying invalid values for a lease or not specifying
+// mandatory lease parameters causes an error while parsing the lease.
+TEST(Lease6Test, fromElementInvalidValues) {
+    // Create valid configuration. We use it as a base from which we will
+    // be removing some of the parameters and some values will be selectively
+    // modified.
+    std::string json = "{"
+        "\"cltt\": 12345678,"
+        "\"duid\": \"00:01:02:03:04:05:06:0a:0b:0c:0d:0e:0f\","
+        "\"fqdn-fwd\": false,"
+        "\"fqdn-rev\": false,"
+        "\"hostname\": \"urania.example.org\","
+        "\"hw-address\": \"08:00:2b:02:3f:4e\","
+        "\"iaid\": 123456,"
+        "\"ip-address\": \"3000::\","
+        "\"preferred-lft\": 400,"
+        "\"prefix-len\": 32,"
+        "\"state\": 0,"
+        "\"subnet-id\": 1234,"
+        "\"type\": \"IA_PD\","
+        "\"valid-lft\": 600"
+        "}";
+
+    // Test invalid parameter values and missing parameters.
+    testInvalidElement<Lease6>(json, "cltt", std::string("xyz"));
+    testInvalidElement<Lease6>(json, "cltt", -1, false);
+    testInvalidElement<Lease6>(json, "duid", "01::00::");
+    testInvalidElement<Lease6>(json, "duid", 1234, false);
+    testInvalidElement<Lease6>(json, "fqdn-fwd", 123);
+    testInvalidElement<Lease6>(json, "fqdn-rev", std::string("foo"));
+    testInvalidElement<Lease6, bool>(json, "hostname", true);
+    testInvalidElement<Lease6>(json, "hw-address", "01::00::", false);
+    testInvalidElement<Lease6>(json, "hw-address", 1234, false);
+    testInvalidElement<Lease6>(json, "iaid", std::string("1234"));
+    testInvalidElement<Lease6>(json, "iaid", -1);
+    testInvalidElement<Lease6, long int>(json, "ip-address", 0xFF000201);
+    testInvalidElement<Lease6>(json, "ip-address", "192.0.2.0", false);
+    testInvalidElement<Lease6>(json, "preferred-lft", std::string("1234"));
+    testInvalidElement<Lease6>(json, "preferred-lft", -1, false);
+    testInvalidElement<Lease6>(json, "prefix-len", std::string("1234"));
+    testInvalidElement<Lease6>(json, "prefix-len", 130);
+    testInvalidElement<Lease6>(json, "state", std::string("xyz"));
+    testInvalidElement<Lease6>(json, "state", 1234, false);
+    testInvalidElement<Lease6>(json, "subnet-id", std::string("xyz"));
+    testInvalidElement<Lease6>(json, "subnet-id", -5, false);
+    testInvalidElement<Lease6>(json, "type", std::string("IA_XY"));
+    testInvalidElement<Lease6>(json, "type", -3, false);
+    testInvalidElement<Lease6>(json, "valid-lft", std::string("xyz"));
+    testInvalidElement<Lease6>(json, "valid-lft", -3, false);
+    testInvalidElement<Lease6>(json, "user-context", "[ ]", false);
+    testInvalidElement<Lease6>(json, "user-context", 1234, false);
+    testInvalidElement<Lease6>(json, "user-context", false, false);
+    testInvalidElement<Lease6>(json, "user-context", "foo", false);
+}
+
 // Verify that the lease states are correctly returned in the textual format.
 TEST(Lease6Test, stateToText) {
     EXPECT_EQ("default", Lease6::statesToText(Lease::STATE_DEFAULT));
