@@ -1,9 +1,11 @@
-// Copyright (C) 2009-2015 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2009-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
 // file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
+#include <config.h>
+
 #include <cctype>
 #include <cassert>
 #include <iterator>
@@ -507,6 +509,12 @@ Name::toText(bool omit_final_dot) const {
     return (ls.toText(omit_final_dot));
 }
 
+std::string
+Name::toRawText(bool omit_final_dot) const {
+    LabelSequence ls(*this);
+    return (ls.toRawText(omit_final_dot));
+}
+
 NameComparisonResult
 Name::compare(const Name& other) const {
     const LabelSequence ls1(*this);
