@@ -1,9 +1,11 @@
-// Copyright (C) 2011-2016 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2011-2018 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
 // file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
+#include <config.h>
+
 #include <exceptions/exceptions.h>
 #include <util/strutil.h>
 #include <util/encode/hex.h>
@@ -143,6 +145,42 @@ TEST(StringUtilTest, Tokens) {
     EXPECT_EQ(string("gamma"), result[3]);
     EXPECT_EQ(string("delta"), result[4]);
     EXPECT_EQ(string("epsilon"), result[5]);
+
+    // Escaped delimiter
+    result = isc::util::str::tokens("foo\\,bar", ",", true);
+    EXPECT_EQ(1, result.size());
+    EXPECT_EQ(string("foo,bar"), result[0]);
+
+    // Escaped escape
+    result = isc::util::str::tokens("foo\\\\,bar", ",", true);
+    ASSERT_EQ(2, result.size());
+    EXPECT_EQ(string("foo\\"), result[0]);
+    EXPECT_EQ(string("bar"), result[1]);
+
+    // Double escapes
+    result = isc::util::str::tokens("foo\\\\\\\\,\\bar", ",", true);
+    ASSERT_EQ(2, result.size());
+    EXPECT_EQ(string("foo\\\\"), result[0]);
+    EXPECT_EQ(string("\\bar"), result[1]);
+
+    // Escaped standard character
+    result = isc::util::str::tokens("fo\\o,bar", ",", true);
+    ASSERT_EQ(2, result.size());
+    EXPECT_EQ(string("fo\\o"), result[0]);
+    EXPECT_EQ(string("bar"), result[1]);
+
+    // Escape at the end
+    result = isc::util::str::tokens("foo,bar\\", ",", true);
+    ASSERT_EQ(2, result.size());
+    EXPECT_EQ(string("foo"), result[0]);
+    EXPECT_EQ(string("bar\\"), result[1]);
+
+    // Escape opening a token
+    result = isc::util::str::tokens("foo,\\,,bar", ",", true);
+    ASSERT_EQ(3, result.size());
+    EXPECT_EQ(string("foo"), result[0]);
+    EXPECT_EQ(string(","), result[1]);
+    EXPECT_EQ(string("bar"), result[2]);
 }
 
 // Changing case
@@ -405,7 +443,7 @@ TEST(StringUtilTest, decodeFormattedHexString) {
     testFormatted("", "");
 
     std::vector<uint8_t> decoded;
-    // Whitepspace.
+    // Whitespace.
     EXPECT_THROW(decodeFormattedHexString("0a ", decoded),
                  isc::BadValue);
     // Whitespace within a string.
@@ -425,4 +463,81 @@ TEST(StringUtilTest, decodeFormattedHexString) {
                  isc::BadValue);
 }
 
+/// @brief Function used to test StringSantitizer
+/// @param original - string to sanitize
+/// @param char_set - regular expression string describing invalid
+/// characters
+/// @param char_replacement - character(s) which replace invalid
+/// characters
+/// @param expected - expected sanitized string
+void sanitizeStringTest(
+    const std::string& original,
+    const std::string& char_set,
+    const std::string& char_replacement,
+    const std::string& expected) {
+
+    StringSanitizerPtr ss;
+    std::string sanitized;
+
+    try {
+        ss.reset(new StringSanitizer(char_set, char_replacement));
+    } catch (const std::exception& ex) {
+        ADD_FAILURE() << "Could not construct sanitizer:" << ex.what();
+        return;
+    }
+
+    try {
+        sanitized = ss->scrub(original);
+    } catch (const std::exception& ex) {
+        ADD_FAILURE() << "Could not scrub string:" << ex.what();
+        return;
+    }
+
+    EXPECT_EQ(sanitized, expected);
+}
+
+// Verifies StringSantizer class
+TEST(StringUtilTest, stringSanitizer) {
+
+    // Bad regular expression should throw.
+    StringSanitizerPtr ss;
+    ASSERT_THROW (ss.reset(new StringSanitizer("[bogus-regex","")), BadValue);
+
+    // List of invalid chars should work: (b,c,2 are invalid)
+    sanitizeStringTest("abc.123", "[b-c2]", "*",
+                       "a**.1*3");
+
+    // Inverted list of valid chars should work: (b,c,2 are valid)
+    sanitizeStringTest("abc.123", "[^b-c2]", "*",
+                       "*bc**2*");
+
+    // A string of all valid chars should return an identical string.
+    sanitizeStringTest("-_A--B__Cabc34567_-", "[^A-Ca-c3-7_-]", "x",
+                       "-_A--B__Cabc34567_-");
+
+    // Replacing with a character should work.
+    sanitizeStringTest("A[b]c\12JoE3-_x!B$Y#e", "[^A-Za-z0-9_]", "*",
+                       "A*b*c*JoE3*_x*B*Y*e");
+
+    // Removing (i.e.replacing with an "empty" string) should work.
+    sanitizeStringTest("A[b]c\12JoE3-_x!B$Y#e", "[^A-Za-z0-9_]", "",
+                       "AbcJoE3_xBYe");
+
+    // More than one non-matching in a row should work.
+    sanitizeStringTest("%%A%%B%%C%%", "[^A-Za-z0-9_]", "x",
+                       "xxAxxBxxCxx");
+
+    // Removing more than one non-matching in a row should work.
+    sanitizeStringTest("%%A%%B%%C%%", "[^A-Za-z0-9_]", "",
+                       "ABC");
+
+    // Replacing with a string should work.
+    sanitizeStringTest("%%A%%B%%C%%", "[^A-Za-z0-9_]", "xyz",
+                       "xyzxyzAxyzxyzBxyzxyzCxyzxyz");
+
+    // Dots as valid chars work.
+    sanitizeStringTest("abc.123", "[^A-Za-z0-9_.]", "*",
+                       "abc.123");
+}
+
 } // end of anonymous namespace
